#!/usr/bin/env python3

"""The core app entrypoint."""

from datetime import datetime
import logging
from pathlib import Path
import tomllib

import cherrypy
import click
from flask import Flask, request
from tentacles.blueprints import (
    api,
    file_ui,
    job_ui,
    printer_ui,
    user_ui,
)
from tentacles.globals import _ctx, Ctx, ctx
from tentacles.store import Store
from tentacles.workers import Worker


@click.group()
def cli():
    pass


def db_factory(app):
    store = Store(app.config.get("db", {}).get("uri"))
    store.connect()
    return store


def custom_ctx(app, wsgi_app):
    def helper(environ, start_response):
        store = db_factory(app)
        token = _ctx.set(Ctx(store))
        try:
            return wsgi_app(environ, start_response)
        finally:
            _ctx.reset(token)
            store.close()

    return helper


def create_j2_request_global(app):
    app.jinja_env.globals["ctx"] = ctx
    app.jinja_env.globals["request"] = request
    app.jinja_env.globals["datetime"] = datetime


def user_session():
    if (
        (
            (session_id := request.cookies.get("sid", ""))
            and (uid := ctx.db.try_key(session_id))
        )
        or (
            request.authorization
            and request.authorization.token
            and (uid := ctx.db.try_key(request.authorization.token))
        )
        or (
            (api_key := request.headers.get("x-api-key"))
            and (uid := ctx.db.try_key(api_key))
        )
    ):
        ctx.sid = session_id
        ctx.uid = uid
        user = ctx.db.fetch_user(uid)
        ctx.gid = user.group_id
        ctx.username = user.name
        ctx.is_admin = user.group_id == 0


def make_app():
    app = Flask(__name__)

    # Before first request
    create_j2_request_global(app)

    # Before request
    app.before_request(user_session)

    # Blueprints
    app.register_blueprint(user_ui.BLUEPRINT)
    app.register_blueprint(printer_ui.BLUEPRINT)
    app.register_blueprint(job_ui.BLUEPRINT)
    app.register_blueprint(file_ui.BLUEPRINT)
    app.register_blueprint(api.BLUEPRINT)

    # Shove our middleware in there
    app.wsgi_app = custom_ctx(app, app.wsgi_app)

    return app


@cli.command()
@click.option("--hostname", "hostname", type=str, default="0.0.0.0")
@click.option("--port", "port", type=int, default=8080)
@click.option("--config", type=Path)
def serve(hostname: str, port: int, config: Path):
    logging.basicConfig(
        format="%(asctime)s %(relativeCreated)6d %(threadName)s - %(name)s - %(levelname)s -  %(message)s",
        level=logging.INFO,
    )

    logging.getLogger("tentacles").setLevel(logging.DEBUG)

    app = make_app()

    if config:
        with open(config, "rb") as fp:
            app.config.update(tomllib.load(fp))
    print(app.config)

    cherrypy.server.unsubscribe()
    server = cherrypy._cpserver.Server()
    cherrypy.config.update(
        {
            "environment": "production",
            "engine.autoreload.on": False,
        }
    )
    cherrypy.tree.graft(app, "/")

    server.socket_host = hostname
    server.socket_port = port
    server.thread_pool = 16
    server.shutdown_timeout = 1
    server.subscribe()

    # Spawn the worker thread
    Worker(cherrypy.engine, app, db_factory, frequency=5).start()

    # Run the server
    cherrypy.engine.start()
    cherrypy.engine.block()


if __name__ == "__main__":
    cli()
